import numpy as np

import logging as log

from ..qt_compat import QtGui, QtCore, QtWidgets

from ..widgets import HWidgets, PLineEdit, TComboBox, \
    SubHeaderLabel,MultiSourceCombo
from ..core import DataModel, LayerManager, LabelManager, Launcher

class Plugin(QtWidgets.QScrollArea):

    name = "Plugin"
    header = None

    Widget = 'widget'
    Plugin = 'plugin'

    completed = QtCore.pyqtSignal(bool)

    def __init__(self, ptype=Plugin, height=0, width=300, dock='left', parent=None):

        super(Plugin, self).__init__(parent=parent)

        self.ptype = ptype
        self.container = QtWidgets.QWidget()
        self.container.setWindowTitle(self.name)
        self.setWidget(self.container)
        self.setWidgetResizable(True)

        self.layout = QtWidgets.QGridLayout(self.container)
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        if ptype == self.Plugin:
            self.layout.setContentsMargins(0,10,0,0)
        self.container.setLayout(self.layout)
        self.row = 0
        self.tab_idx = 0

    def setup(self):
        self.header.setEnabled(True)
        self.header.setVisible(True)

    def addWidget(self, widget, align=QtCore.Qt.AlignTop):
        self.layout.addWidget(widget, self.row, 0, align)
        self.row += 1

    def __add__(self, widget):
        self.addWidget(widget)
        return self

    def toggle(self):
        self.setVisible(not self.isVisible())

    def value(self):
        pass

    def on_focus(self):
        pass

    def on_tab_changed(self, idx):
        pass


class PredWidgetBase(QtWidgets.QWidget):
    """
    Base class for the Predict and TrainPredict Widgets
    """

    def __init__(self, parent=None):
        super(PredWidgetBase, self).__init__(parent=None)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()
        self.selected_level = -1
        self.parent_level = None
        self.parent_label = -1

        self.vbox = QtWidgets.QVBoxLayout()
        self.setLayout(self.vbox)

        self.vbox.addWidget(SubHeaderLabel('Descriptor'))

        # Region Type
        self.use_region = TComboBox('Region:', ['Voxels', 'SuperVoxels'], selected=1)
        self.use_region.currentIndexChanged.connect(self.on_region_changed)
        self.use_region.setMinimumWidth(200)

        # Descriptor type
        self.use_desc = MultiSourceCombo()

        self.desc_type = TComboBox('Type:',
                                   ['Mean', 'Quantized', 'Textons',
                                    'Covar', 'Sigma Set'])
        self.desc_bins = PLineEdit(10, parse=int)
        self.desc_bins.setMaximumWidth(60)

        self.nh_order = TComboBox('Order:', [0, 1, 2], selected=0)

        self.preprocess = TComboBox('Project:',
                                    ['None', 'standard', 'random_projection',
                                     'pca', 'random_pca'])

        self.vbox.addWidget(HWidgets(None, self.use_region, stretch=[1, 0]))
        self.vbox.addWidget(HWidgets(None, 'Features:', self.use_desc, stretch=[1, 0, 0]))
        self.supervoxel_desc_params = HWidgets(self.desc_type, 'Bins:',
                                               self.desc_bins, self.nh_order,
                                               stretch=[1, 0, 0, 1])
        self.vbox.addWidget(self.supervoxel_desc_params)
        self.vbox.addWidget(HWidgets(None, self.preprocess, stretch=[1, 0, 0]))

        self.vbox.addWidget(SubHeaderLabel('Prediction Refinement'))
        self.refinement_combo = TComboBox('Refine:', ['None', 'Potts',
                                                      'Appearance', 'Edge'])
        self.refinement_lamda = PLineEdit(50., parse=float)
        self.refinement_lamda.setMaximumWidth(80)
        self.refinement_desc = MultiSourceCombo()
        self.ref_features = HWidgets(None, 'Features:', self.refinement_desc,
                                     stretch=[1, 0, 0])
        self.ref_features.setVisible(False)
        self.vbox.addWidget(HWidgets(self.refinement_combo, 'Lambda:', self.refinement_lamda))
        self.vbox.addWidget(self.ref_features)

        self.refinement_combo.currentIndexChanged.connect(self.on_ref_type_changed)
        self.use_region.currentIndexChanged.connect(self.on_region_changed)
        self.DM.voxel_descriptor_computed.connect(self.on_voxel_desc_changed)
        self.DM.voxel_descriptor_removed.connect(self.on_voxel_desc_changed)
        self.DM.supervoxel_descriptor_computed.connect(self.on_voxel_desc_changed)
        self.DM.supervoxel_descriptor_computed.connect(self.on_voxel_desc_changed)
        self.DM.supervoxel_descriptor_cleared.connect(self.on_voxel_desc_changed)

    def select_level(self, level, plevel, plabel):
        self.selected_level = level
        self.parent_level = plevel
        self.parent_label = plabel

    def on_voxel_desc_changed(self):
        self.on_region_changed(self.use_region.currentIndex())

    def on_ref_type_changed(self, idx):
        self.ref_features.setVisible(idx > 2)

    def on_region_changed(self, idx):
        self.supervoxel_desc_params.setVisible(idx == 1)

    def on_predict(self, params):
        ### LEVEL AND LABEL
        level = self.selected_level

        if level < 0 or level is None:
            QtWidgets.QMessageBox.critical(self, "Error", "No level selected")
            return

        parent_level = self.parent_level
        parent_label = self.parent_label
        idxs = self.LBLM.idxs(level=level)

        if len(idxs) == 0:
            self.launcher.error.emit('Current level doesn\'t contain any label')
            return

        level_params = dict(level=level, plevel=parent_level, plabel=parent_label)

        ### DESCRIPTOR
        desc_params = dict()
        desc_params['supervoxels'] = 'supervoxels/supervoxels' if self.use_region.currentIndex() == 1 else None
        desc_params['features'] = self.use_desc.value()
        desc_params['projection'] = self.preprocess.value()
        if desc_params['supervoxels']:
            desc_params['desc_type'] = self.desc_type.value()
            desc_params['desc_bins'] = self.desc_bins.value()
            desc_params['nh_order'] = self.nh_order.value()
            desc_params['sp_edges'] = 'supervoxels/graph_edges'

        if len(desc_params['features']) == 0:
            self.launcher.error.emit('No descriptor feature has been selected.')
            return

        ### CLASSIFIER
        clf_params = params
        self.DM.add_desc_params_to_model(desc_params)

        ### REFINEMENT
        ref_params = dict()
        ref_params['ref_type'] = self.refinement_combo.value()
        ref_params['lambda'] = self.refinement_lamda.value()
        ref_params['features'] = self.refinement_desc.value()
        ref_params['sp_edges'] = 'supervoxels/graph_edges'
        ref_params['sp_eweights'] = 'supervoxels/graph_edge_weights'

        if self.refinement_combo.currentIndex() > 2 and len(ref_params['features']) == 0:
            self.launcher.error.emit('No refinement feature has been selected.')
            return

            ### ANNOTATIONS
        y_data = self.LBLM.dataset(level)
        out_labels = 'predictions/predictions'
        out_confidence = 'predictions/confidence'

        if parent_level != None and parent_level >= 0 and parent_label >= 0:
            p_data = self.LBLM.dataset(parent_level)
        else:
            p_data = None

        self.DM.remove_dataset(out_labels)
        self.DM.remove_dataset(out_confidence)

        dataset = 'predictions/predictions'
        self.DM.create_empty_dataset(out_labels, self.DM.data_shape,
                                     np.int16, check=False)
        self.DM.create_empty_dataset(out_confidence, self.DM.data_shape,
                                     np.float32, check=False)

        log.info("Level {}".format(level_params))
        log.info("Desc {}".format(desc_params))
        log.info("Clf {}".format(clf_params))
        log.info("Ref {}".format(ref_params))
        log.info("{} {}".format(y_data, p_data))
        log.info("{} {}".format(out_labels, out_confidence))

        self.run_prediction(y_data, p_data, level_params, desc_params, ref_params,
                            clf_params, out_labels, out_confidence, level)

    def run_prediction(self, y_data, p_data, level_params, desc_params, ref_params,
                       clf_params, out_labels, out_confidence, level):
        """
        To be implemented in subclasses
        """
        raise NotImplementedError

    def on_predicted(self, params):
        predictions, uncertainty, labels = params
        if not self.LM.isin('Predictions', 'Predictions'):
            self.LM.addLayer(predictions, 'Predictions', 'Predictions', alpha=0.5, visible=True)
        self.DM.level_predicted.emit(self.selected_level)




import numpy as np
import time
from sklearn.cluster import MiniBatchKMeans, KMeans, Birch
from scipy.ndimage import gaussian_filter
from scipy.ndimage.measurements import label as cclabel
from skimage.util import img_as_float
from skimage import color
from skimage.segmentation import relabel_sequential
from skimage.segmentation._slic import _enforce_label_connectivity_cython

from .rag import rag_from_neighbors
from .spencoding import spmeans
from ._qpbo import solve_binary, solve_aexpansion
from ._supersegments import _neighbors, _postprocess_labels

class Timer(object):
    def __init__(self, name='Timer'):
        self.name = name

    def __enter__(self):
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        self.tend = (time.time() - self.tstart)
        print('[%s] Elapsed: %.4f seconds' % (self.name, self.tend))

def mean_hue(h1, h2):
    mh = np.zeros_like(h1)
    mask1 = np.abs(h2 - h1) > 180
    mask2 = mask1 & (h1 + h2 < 360.)
    mask3 = mask1 & (h1 + h2 >= 360.)
    mh[mask2] = (h1[mask2] + h2[mask2] + 360.) / 2.
    mh[mask3] = (h1[mask3] + h2[mask3] - 360.) / 2.
    mh[~mask1] = (h1[~mask1] + h2[~mask1]) / 2.
    return mh

def colordiff(A, B):
    if B.shape[0] == 1:
        B = np.repeat(B, A.shape[0], axis=0)
    ac = np.sqrt(A[:, 1]**2 + A[:, 2]**2)
    bc = np.sqrt(B[:, 1]**2 + B[:, 2]**2)
    mc = (ac + bc)/2
    g = (1 - np.sqrt(mc**7 / (mc**7 + 25.0**7))) / 2.
    A = A.copy()
    A[:, 1] *= (1 + g)
    B = B.copy()
    B[:, 1] *= (1 + g)

    A = color.lab2lch(A)
    B = color.lab2lch(B)

    dl = (B[:, 0] - A[:, 0])
    dc = (B[:, 1] - A[:, 1])
    dh = (B[:, 2] - A[:, 2])
    mask1 = (A[:, 1] * B[:, 1] == 0)
    mask2 = (~mask1) & (dh > 180)
    mask3 = (~mask1) & (dh < -180)
    dh[mask1] = 0
    dh[mask2] -= 360
    dh[mask3] += 360

    dh = 2 * np.sqrt(A[:, 1] * B[:, 1]) * np.sin(np.deg2rad(dh / 2.))
    ml = (A[:, 0] + B[:, 0]) / 2
    mc = (A[:, 1] + B[:, 1]) / 2
    mh = np.zeros_like(dh)
    mh[mask1] = A[mask1, 2] + B[mask1, 2]
    mh[~mask1] = mean_hue(A[~mask1, 2], B[~mask1, 2])

    mls = (ml - 50)**2
    sl = 1.0 + 0.015 * mls / np.sqrt(20 + mls)

    # chroma weight
    sc = 1 + 0.045 * mc

    # hue weight
    t = 1 - 0.17 * np.cos(np.deg2rad(mh - 30)) + \
            0.24 * np.cos(np.deg2rad(2 * mh)) + \
            0.32 * np.cos(np.deg2rad(3 * mh + 6)) - \
            0.20 * np.cos(np.deg2rad(4 * mh - 63))
    sh = 1 + 0.015 * mc * t

     # rotation term
    dtheta = 30 * np.exp(-((mh - 275)/25)**2)
    cr = 2 * np.sqrt(mc**7 / (mc**7 + 25.0**7))
    tr = -np.sin(np.deg2rad(2*dtheta)) * cr
    return np.sqrt((dl/(1*sl))**2 + (dc/(1*sc))**2 + (dh/(1*sh))**2 + \
                   tr * (dc/(1*sc)) * (dh/(1*sh)))




def HPC(data, edges, nbow=20, lamda=1, sampling='random', nsamples=10000,
        label_potential='l1', unary_sq=True, online=True, gamma=None, max_iter=5,
        truncated=False, rng=42, return_centers=False, max_size=1000, min_size=10,
        max_neighbors=50, verbose=True, return_image=False, num_clusters=None,
        raw_data_group=False, colordiff=False):

    labels, label_cost, centers = qmrf_regions(data, edges, nbow=nbow, lamda=lamda, sampling='random', nsamples=nsamples,
                                               label_potential=label_potential, unary_sq=unary_sq, online=online,
                                               max_iter=max_iter, gamma=gamma, return_centers=True,
                                               verbose=verbose)

    splabels = postprocess_regions(labels, edges, label_cost, data, max_size=max_size,
                                   min_size=min_size, max_neighbors=max_neighbors,
                                   num_clusters=num_clusters, verbose=verbose,
                                   raw_data_group=raw_data_group)
    if return_centers:
        return splabels, labels, centers
    return splabels


def img2data(image, lab=True, sigma=None):
    if image.shape[-1] in (3, 4) and lab:
        image = color.rgb2lab(image)
    image = img_as_float(image)
    if sigma is not None:
        image = gaussian_filter(image, sigma)

    if image.shape[-1] in (3, 4):
        return image.reshape(-1, image.shape[-1])
    else:
        return image.reshape(-1, 1)


def grid_edges(shape, connectivity=4):
    shape = shape if type(shape) == tuple else shape.shape
    B = np.arange(np.prod(shape[:2]), dtype=np.int32).reshape(shape[:2])
    edges_h = np.dstack([B[:, :-1], B[:, 1:]]).reshape(-1, 2)
    edges_v = np.dstack([B[:-1, :], B[1:, :]]).reshape(-1, 2)
    edges = np.r_[edges_h, edges_v]
    if connectivity == 8:
        edges_ht = np.dstack([B[1:, :-1], B[:-1, 1:]]).reshape(-1, 2)
        edges_hb = np.dstack([B[:-1, :-1], B[1:, 1:]]).reshape(-1, 2)
        edges = np.r_[edges, edges_ht, edges_hb]
    return edges


def cube_edges(image, connectivity=6):
    B = np.arange(np.prod(image.shape[:3]), dtype=np.int32).reshape(image.shape[:3])
    edges_h = np.dstack([B[:, :, :-1], B[:, :, 1:]]).reshape(-1, 2)
    edges_v = np.dstack([B[:, :-1, :], B[:, 1:, :]]).reshape(-1, 2)
    edges_d = np.dstack([B[:-1, :, :], B[1:, :, :]]).reshape(-1, 2)
    edges = np.r_[edges_h, edges_v, edges_d]
    if connectivity == 18:
        edges_1 = np.dstack([B[:, :-1, :-1], B[:, 1:, 1:]]).reshape(-1, 2)
        edges_2 = np.dstack([B[:, 1:, :-1], B[:, :-1, 1:]]).reshape(-1, 2)
        edges_3 = np.dstack([B[:-1, :, :-1], B[1:, :, 1:]]).reshape(-1, 2)
        edges_4 = np.dstack([B[1:, :, :-1], B[:-1, :, 1:]]).reshape(-1, 2)
        edges_5 = np.dstack([B[:-1, :-1, :], B[1:, 1:, :]]).reshape(-1, 2)
        edges_6 = np.dstack([B[1:, :-1, :], B[:-1, 1:, :]]).reshape(-1, 2)
        edges = np.r_[edges, edges_1, edges_2, edges_3, edges_4, edges_5, edges_6]
    if connectivity == 24:
        edges_1 = np.dstack([B[:-1, :-1, :-1], B[1:, 1:, 1:]]).reshape(-1, 2)
        edges_2 = np.dstack([B[1:, :-1, :-1], B[:-1, 1:, 1:]]).reshape(-1, 2)
        edges_3 = np.dstack([B[:-1, 1:, :-1], B[1:, :-1, 1:]]).reshape(-1, 2)
        edges_4 = np.dstack([B[:-1, :-1, 1:], B[1:, 1:, :-1]]).reshape(-1, 2)
        edges = np.r_[edges, edges_1, edges_2, edges_3, edges_4]
    return edges


def qmrf_regions(data, edges, nbow=20, lamda=1, sampling='random', nsamples=10000,
                 label_potential='l1', unary_sq=True, online=True, gamma=None, max_iter=5,
                 truncated=False, rng=42, verbose=True, return_centers=False,
                 return_edge_costs=True):
    with Timer('Colors'):
        if nbow == 'birch':
            clf = Birch(threshold=0.8, branching_factor=100)
        elif online:
            clf = MiniBatchKMeans(n_clusters=nbow, verbose=verbose, random_state=rng,
                                  batch_size=100, max_iter=100, max_no_improvement=10)
        else:
            clf = KMeans(n_clusters=nbow, verbose=verbose, random_state=rng)

        if nsamples is None:
            dist = clf.fit_transform(data)
        else:
            if sampling == 'random':
                idx = np.random.choice(data.shape[0], nsamples, replace=False)
            else:
                n = np.sqrt(nsamples)
                ratio = image.shape[0] / float(image.shape[1])
                ny = int(n * ratio)
                nx = int(n / ratio)
                y = np.linspace(0, image.shape[0], ny, endpoint=False) + (image.shape[0]//ny//2)
                x = np.linspace(0, image.shape[1], nx, endpoint=False) + (image.shape[1]//nx//2)
                xx, yy = np.meshgrid(x, y)
                idx = np.round(yy * image.shape[1] + xx).astype(int).flatten()
            clf.fit(data[idx])
            dist = clf.transform(data)

        if nbow == 'birch':
            centers = clf.subcluster_centers_
        else:
            centers = clf.cluster_centers_

    with Timer('Unary'):
        K = centers.shape[0]

        if label_potential == 'color':
            unary_cost = np.zeros((data.shape[0], centers.shape[0]), np.float32)
            for i in range(centers.shape[0]):
                unary_cost[:, i] = colordiff(data, centers[i:i+1])
        else:
            unary_cost = dist.astype(np.float32)

        if unary_sq:
            unary_cost **= 2

    with Timer('Pairwise'):
        if label_potential == 'l1':
            label_cost = np.abs(centers[:, None, :] - centers[None, ...]).sum(-1)
        elif label_potential == 'l2':
            label_cost = np.sqrt(((centers[:, None, :] - centers[None, ...])**2).sum(-1))
        elif label_potential == 'potts':
            label_cost = np.ones((K, K), int) - np.eye(K, dtype=int)
        elif label_potential == 'color':
            label_cost = np.zeros((centers.shape[0], centers.shape[0]), np.float32)
            for i in range(centers.shape[0]):
                label_cost[:, i] = colordiff(centers, centers[i:i+1])
        if truncated:
            label_cost = np.maximum(1, label_cost)
        label_cost = (label_cost * lamda).astype(np.float32)

    if verbose:
        print("=================")
        print("Minimizing graph:")
        print("Nodes: %d, edges: %d, labels: %d" % \
              (unary_cost.shape[0], edges.shape[0], label_cost.shape[0]))
        print("UnarySq: %s, LabelPotential: %s, EdgeCost: %s" % \
              (unary_sq, label_potential, (gamma is not None)))
        print("#################")

    with Timer('Edge Cost'):
        diff = ((data[edges[:, 0]] - data[edges[:, 1]])**2).sum(axis=1)
        if gamma is not None and type(gamma) in [int, float]:
            edge_costs = np.exp(-gamma * diff).astype(np.float32)
        elif gamma == 'auto':
            edge_costs = np.exp(-diff.mean() * diff).astype(np.float32)
        elif gamma == 'color':
            edge_costs = 1. / (1. + colordiff(data[edges[:, 0]], data[edges[:, 1]]))
            edge_costs = edge_costs.astype(np.float32)
        else:
            edge_costs = np.ones(edges.shape[0], dtype=np.float32)

    with Timer('Minimize'):
        if label_cost.shape[0] == 2:
            labels = solve_binary(edges, unary_cost, edge_costs, label_cost)
        else:
            labels = solve_aexpansion(edges, unary_cost, edge_costs, label_cost)

    if return_centers:
        return labels, label_cost, centers

    return labels, label_cost


def postprocess_regions(labels, edges, label_cost, data, max_size=1000, min_size=None, max_neighbors=50,
                        num_clusters=None, verbose=True, raw_data_group=False):
    neighbors = _neighbors(edges, labels.shape[0], max_neighbors)
    splabels = _postprocess_labels(labels, neighbors, max_size=max_size)

    num_sps = splabels.max()+1

    print("Num Superpixels: %d" % (splabels.max()+1))

    if min_size is not None and min_size > 1:
        counts = np.bincount(splabels)
        rag = rag_from_neighbors(splabels, neighbors)

        for p, d in rag.nodes(data=True):
            d['labels'] = {p}
            d['olabel'] = p

        if verbose:
            print("Num smaller superpixels: %d" % (counts < min_size).sum())

        spedges = np.array(rag.edges(), int)
        if not raw_data_group:
            olabels = spmeans(labels[:, None].astype(np.float32), splabels,
                              splabels.max()+1).astype(np.int32)
            edge_cost = np.ravel(label_cost[olabels[spedges[:, 0]],
                                            olabels[spedges[:, 1]]])
        else:
            costs = spmeans(data.astype(np.float32), splabels, splabels.max()+1)
            edge_cost = ((costs[spedges[:, 0]] - costs[spedges[:, 1]])**2).sum(1)
            edge_cost = edge_cost.astype(np.float32)

        if raw_data_group:
            order = edge_cost + (min(counts[spedges[:, 0]] * counts[spedges[:, 1]]))/ counts.max()
        else:
            order = edge_cost
        idx = np.ravel(np.argsort(order))

        for i in idx:
            p, q = spedges[i]
            if q not in rag.node[p]['labels'] and (counts[p] < min_size or counts[q] < min_size):
                plabels = rag.node[p]['labels']
                qlabels = rag.node[q]['labels']
                pset = list(rag.node[p]['labels'] - set([p, q]))
                qset = list(rag.node[q]['labels'] - set([p, q]))
                counts[pset] += counts[q]
                counts[qset] += counts[p]
                total_nodes = plabels | qlabels
                for i in total_nodes:
                    rag.node[i]['labels'] = total_nodes
                counts[p] += counts[q]
                counts[q] += counts[p]

                num_sps -= 1
                if num_clusters is not None and num_clusters >= num_sps:
                    break

        qlabels = np.arange(len(rag))
        label_map = np.full(qlabels.size, -1, np.int32)
        for ix, (n, d) in enumerate(rag.nodes_iter(data=True)):
            for label in d['labels']:
                if label_map[label] == -1:
                    label_map[label] = ix

        qlabels = label_map[qlabels]
        splabels = qlabels[splabels]

    splabels = relabel_sequential(splabels)[0]
    splabels -= splabels.min()

    if verbose and min_size is not None:
        print("Num Superpixels after merging: %d" % (splabels.max()+1))

    return splabels




from ..qt_compat import QtGui, QtCore, QtWidgets

import six
import os
from collections import defaultdict
from ..core import DataModel, LabelManager

DEFAULT_DIR_KEY = "default_dir"
DEFAULT_DATA_KEY = "default_data_dir"

class SectionCombo(QtWidgets.QToolButton):

    currentIndexChanged = QtCore.pyqtSignal(int)

    def __init__(self, text='Select', parent=None):
        super(SectionCombo, self).__init__(parent)
        self.setMaximumHeight(30)
        self.setStyleSheet('QToolButton {background-color: #D0EAFF; padding-right: 20px;}'
                           'QMenu {background-color: #A5D1F3;}')
        self.setText(text)
        self.toolmenu = QtWidgets.QMenu(self)
        self.toolmenu.setContentsMargins(3,3,3,3)
        self.setMenu(self.toolmenu)
        self.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.setMinimumWidth(100)
        self._names = []
        self._actions = []
        self.selected_idx = 0
        self.item_count = 0

    def setMinimumWidth(self, val):
        super(SectionCombo, self).setMinimumWidth(val)
        self.toolmenu.setMinimumWidth(val)

    def addItem(self, item, section=False):
        curr_item = item
        widget = QtWidgets.QWidgetAction(self.toolmenu)
        if section:
            lbl = QtWidgets.QLabel(item)
            lbl.setStyleSheet('background-color: #fefefe; color: #000;'
                              'padding-top: 4px; padding-bottom: 4px;'
                              'padding-left: 10px; padding-right: 0px;'
                              'font-weight: bold;')
            widget.setEnabled(False)
        else:
            lbl = QtWidgets.QPushButton(item)
            lbl.setStyleSheet('border-radius: 0px; border: 0px;'
                              'text-align: left; padding-left: 15px;')
            lbl.clicked.connect(lambda: self.on_clicked(curr_item))
        widget.setDefaultWidget(lbl)
        self._names.append(item)
        self._actions.append(widget)
        self.toolmenu.addAction(widget)
        self.item_count += 1

    def sizeHint(self):
        size = self.toolmenu.sizeHint()
        size.setWidth(size.width() + 20)
        return size

    def on_clicked(self, item):
        idx = self._names.index(item)
        self.selected_idx = idx
        self.setText(self._names[idx])
        self.toolmenu.close()
        self.currentIndexChanged.emit(idx)

    def setCurrentIndex(self, idx):
        self.selected_idx = idx
        self.setText(self._names[idx])

    def currentIndex(self):
        return self.selected_idx

    def removeItem(self, idx):
        action = self._actions[idx]
        del self._names[idx]
        del self._actions[idx]
        self.toolmenu.removeAction(action)
        if idx == self.selected_idx:
            self.currentIndexChanged.emit(0)
            self.setCurrentIndex(0)

    def clear(self):
        self._names = []
        self._actions = []
        self.toolmenu.clear()


class FileWidget(QtWidgets.QWidget):

    path_updated = QtCore.pyqtSignal(str)

    def __init__(self, extensions='*.h5', home=None, folder=False, save=True, parent=None):
        super(FileWidget, self).__init__(parent)
        hbox = QtWidgets.QHBoxLayout()
        hbox.setContentsMargins(0,0,0,0)
        self.setLayout(hbox)

        self.extensions = extensions
        self.folder = folder
        self.save = save

        if save and home == '~':
            home = os.path.expanduser('~')
        elif home is None:
            home = 'Click to select Folder' if folder else 'Click to select File'

        self.path = QtWidgets.QLineEdit(home)
        self.path.setReadOnly(True)
        self.path.mousePressEvent = self.find_path
        self.selected = False

        hbox.addWidget(self.path)

    def find_path(self, ev):
        if ev.button() != 1:
            return
        self.open_dialog()

    def open_dialog(self):
        selected = False
        path = None
        settings = QtCore.QSettings()
        if self.folder:
            flags = QtWidgets.QFileDialog.ShowDirsOnly
            message = 'Select Output Folder'
            path = QtWidgets.QFileDialog.getExistingDirectory(self, message,
                                                              settings.value(DEFAULT_DIR_KEY), flags)
            if path is not None and len(path) > 0 and os.path.isdir(path):
                selected = True
                settings.setValue(DEFAULT_DIR_KEY, path)
        else:
            folder = settings.value(DEFAULT_DATA_KEY)
            if not folder:
                folder = settings.value(DEFAULT_DIR_KEY)
            if self.save:
                path, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Select input source", folder,
                                                         filter=self.extensions)
            else:
                path, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Select input source", folder,
                                                         filter=self.extensions)
            if path is not None and len(path) > 0:
                selected = True
                settings.setValue(DEFAULT_DATA_KEY, os.path.dirname(path))

        if selected:
            self.path.setText(path)
            self.path_updated.emit(path)
            self.selected = True

    def value(self):
        return self.path.text() if self.selected else None


class ActionButton(QtWidgets.QPushButton):
    def __init__(self, text, parent=None):
        super(ActionButton, self).__init__(text, parent=parent)
        self.setMinimumWidth(70)


class ActionButton(QtWidgets.QPushButton):
    def __init__(self, text, parent=None):
        super(ActionButton, self).__init__(text, parent=parent)
        self.setMinimumWidth(70)


class CheckableCombo(QtWidgets.QToolButton):

    selectionChanged = QtCore.pyqtSignal()

    def __init__(self, text='Select', parent=None):
        super(CheckableCombo, self).__init__(parent)
        self.setMaximumHeight(30)
        self.setText(text)
        self.toolmenu = QtWidgets.QMenu(self)
        self.setMenu(self.toolmenu)
        self.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.setStyleSheet('QToolButton {'
                           '  color: #0056b3; background-color: #D0EAFF;'
                           '  padding-right: 25px; border-radius: 4px;'
                           '  border: 1px solid #fefefe;'
                           '  margin: 0px;'
                           '}'
                           'QMenu {'
                           '  background: #D0EAFF; padding: 10px;'
                           '}'
                           'QCheckBox {color: #0056b3;}'
                           'QAction{ padding: 5px;}')
        self.setMinimumWidth(100)
        self._names = []
        self._data = []

    def sizeHint(self):
        size = self.toolmenu.sizeHint()
        size.setWidth(size.width() + 20)
        return size

    def addItem(self, item):
        chk = QtWidgets.QCheckBox(item + '    ')
        chk.stateChanged.connect(self.on_selection_changed)
        widget = QtWidgets.QWidgetAction(self.toolmenu)
        widget.setDefaultWidget(chk)
        self.toolmenu.addAction(widget)
        self._names.append(item)
        self._data.append((widget, chk))
        self.toolmenu.repaint()
        self.repaint()

    def on_selection_changed(self, *args):
        self.selectionChanged.emit()

    def setItemText(self, item, text):
        self._data[item][1].setText(text + '    ')
        self.toolmenu.repaint()
        self.repaint()

    def removeItem(self, idx):
        self.toolmenu.removeAction(self._data[idx][0])
        del self._data[idx]
        del self._names[idx]

    def clear(self):
        self._names = []
        self._data = []
        self.toolmenu.clear()
        self.toolmenu.repaint()
        self.repaint()

    def getSelectedIndexes(self):
        return [i for i in range(len(self._data))
                if self._data[i][1].isChecked()]

    def checkGivenItems(self, item_list):
        """
        Given a list of items, checks those items in the CheckableCombo

        :param item_list: List of strings. Labels of boxes to be checked 
        """
        for i in range(len(self._data)):
            # Have to truncate the string here as 4 extra spaces were added in AddItem
            if self._data[i][1].text()[:-4] in item_list:
                self._data[i][1].setChecked(True)

class CheckableLabels(CheckableCombo):
    def __init__(self, restrict_level=None, parent=None):
        super(CheckableLabels, self).__init__('Select Labels', parent)
        self.DM = DataModel.instance()
        self.LBLM = LabelManager.instance()

        self.restrict_level = restrict_level
        self._labels = []
        self.LBLM.labelAdded.connect(self.on_label_added)
        self.LBLM.labelLoaded.connect(self.on_label_added)
        self.LBLM.labelRemoved.connect(self.on_label_removed)
        self.LBLM.labelNameChanged.connect(self.on_label_name_changed)

    def selectLevel(self, level):
        if level >= 0:
            self.restrict_level = level
            self._labels = []
            self.clear()
            for lbl in self.LBLM.labels(level):
                self.on_label_added(level, '', lbl.idx, lbl.name)

    def on_label_added(self, level, dataset, label, name, *args):
        if self.restrict_level is None:
            self._labels.append((level, label))
            self.addItem('Level {}/{}'.format(level, name))
        elif level == self.restrict_level:
            self._labels.append((level, label))
            self.addItem(name)

    def on_label_removed(self, level, dataset, labelobj):
        if self.restrict_level is None or level == self.restrict_level:
            if (level, labelobj.idx) in self._labels:
                idx = self._labels.index((level, labelobj.idx))
                self.removeItem(idx)
                del self._labels[idx]

    def on_label_name_changed(self, level, dataset, label, newname):
        if self.restrict_level is None:
            idx = self._labels.index((level, label))
            self.setItemText(idx, 'Level {}/{}'.format(level, newname))
        elif level == self.restrict_level:
            idx = self._labels.index((level, label))
            self.setItemText(idx, newname)

    def getSelectedLabels(self):
        return [self._labels[i][1] for i in self.getSelectedIndexes()]


class MultiSourceCombo(CheckableCombo):

    def __init__(self, parent=None):
        super(MultiSourceCombo, self).__init__('Select Sources', parent)
        self.DM = DataModel.instance()
        self.setMinimumWidth(200)
        self.repopulate()

        self.DM.channel_computed.connect(self.on_computed)
        self.DM.channel_removed.connect(self.on_removed)

    def repopulate(self):
        sources = sorted(self.DM.available_channels(return_names=True))
        self.names = ['Data'] + [p['feature_name'] for _, _, p in sources]
        self.sources = ['/data'] + [p for _, p, _ in sources]
        self.clear()
        for source in self.names:
            self.addItem(source)

    def on_update_channel(self, name):
        self.setCurrentIndex(self.sources.index(name))

    def on_computed(self, name, ndict):
        if not name in self.sources:
            self.names.append(ndict['feature_name'])
            self.sources.append(name)
            self.addItem(ndict['feature_name'])

    def on_removed(self, name):
        if name in self.sources:
            idx = self.sources.index(name)
            del self.sources[idx]
            del self.names[idx]
            self.removeItem(idx)

    def value(self):
        return [self.sources[i] for i in self.getSelectedIndexes()]

class SourceCombo(SectionCombo):

    def __init__(self, parent=None, listen_DM=False, listen_self=False,
                 ignore=None):
        super(SourceCombo, self).__init__(parent)
        self.DM = DataModel.instance()
        self.ignore_name = 'channels/{}'.format(ignore) if ignore is not None else None
        self.repopulate()

        self.currentIndexChanged.connect(self.on_selected)
        self.DM.channel_computed.connect(self.on_computed)
        self.DM.channel_removed.connect(self.on_removed)
        if listen_DM:
            self.DM.update_channel.connect(self.on_update_channel)
        if listen_self:
            self.DM.select_channel.connect(self.other_selection)

        #self.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)

    def repopulate(self):
        sources = sorted(self.DM.available_channels(return_names=True))
        self.names = ['Data'] + [p['feature_name'] for _, _, p in sources]
        self.sources = ['/data'] + [p for _, p, _ in sources]
        if self.ignore_name is not None:
            if self.ignore_name in self.sources:
                index = self.sources.index(self.ignore_name)
                del self.names[index]
                del self.sources[index]
        self.clear()
        for source in self.names:
            self.addItem(source)
        self.setCurrentIndex(0)

    def other_selection(self, name):
        self.blockSignals(True)
        self.setCurrentIndex(self.sources.index(name))
        self.blockSignals(False)

    def on_selected(self):
        self.DM.select_channel.emit(self.value())

    def on_update_channel(self, name):
        if name in self.sources:
            self.setCurrentIndex(self.sources.index(name))

    def on_computed(self, name, ndict):
        if not name in self.sources and name != self.ignore_name:
            self.names.append(ndict['feature_name'])
            self.sources.append(name)
            self.addItem(ndict['feature_name'])

    def setIgnore(self, name):
        self.ignore_name = 'channels/{}'.format(name)

    def on_removed(self, name):
        if name == self.ignore_name or name not in self.sources:
            return
        idx = self.sources.index(name)
        del self.sources[idx]
        del self.names[idx]
        self.removeItem(idx)

    def value(self):
        if self.currentIndex() < len(self.sources):
            return self.sources[self.currentIndex()]
        elif len(self.sources) > 0:
            return self.sources[0]
        else:
            return None


    def setSource(self, source):
        self.on_update_channel(source)

    def loadSource(self, source):
        self.blockSignals(True)
        self.on_update_channel(source)
        self.blockSignals(False)

class SComboBox(QtWidgets.QComboBox):
    def __init__(self, *args):
        super(SComboBox, self).__init__(*args)

class ComboDialog(QtWidgets.QDialog):
    def __init__(self, options=None, parent=None):
        super(ComboDialog, self).__init__(parent=parent)

        layout = QtWidgets.QVBoxLayout(self)

        self.combo = SComboBox()
        for option in options:
            self.combo.addItem(option)
        layout.addWidget(self.combo)

        # OK and Cancel buttons
        buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal, self)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    @staticmethod
    def getOption(options, parent=None):
        dialog = ComboDialog(options, parent=parent)
        result = dialog.exec_()
        option = dialog.combo.currentText()
        return (option, result == QtWidgets.QDialog.Accepted)

    @staticmethod
    def getOptionIdx(options, parent=None):
        dialog = ComboDialog(options, parent=parent)
        result = dialog.exec_()
        option = dialog.combo.currentIndex()
        return (option, result == QtWidgets.QDialog.Accepted)

class RoundedWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, color=(168, 34, 3), bg=None, width=2, radius=3):
        super(RoundedWidget, self).__init__(parent=parent)
        self.border_color = color
        self.border_width = width
        self.background = bg
        self.radius = radius
        self.margin = int(width / 2.0)

    def paintEvent(self, e):
        qp = QtGui.QPainter()
        qp.begin(self)

        pen = None
        if type(self.border_color) is tuple:
            pen = QtGui.QPen(QtGui.QColor(*self.border_color))
        elif self.border_color is not None:
            pen = QtGui.QPen(QtGui.QColor(self.border_color))

        brush = None
        if type(self.background) is tuple:
            brush = QtGui.QColor(*self.background)
        elif self.background is not None:
            brush = QtGui.QColor(self.background)

        if pen is not None:
            pen.setWidth(self.border_width)
            qp.setPen(pen)
        else:
            qp.setPen(QtCore.Qt.NoPen)

        if brush is not None:
            qp.setBrush(brush)

        qp.drawRoundedRect(self.margin, self.margin,
                           self.width()-2*self.margin-1,
                           self.height()-2*self.margin-1,
                           self.radius, self.radius);
        qp.end()


class PicButton(QtWidgets.QToolButton):
    def __init__(self, pixmap, parent=None):
        super(PicButton, self).__init__(parent)
        self.icon = QtGui.QIcon(pixmap)
        self.setIcon(self.icon)

    def sizeHint(self):
        return QtCore.QSize(50, 30)


class HeaderLabel(QtWidgets.QLabel):

    def __init__(self, text, parent=None, height=30,
                 bgcolor='#6194BC', color='#fefefe', bradius=0,
                 bsides=0, fontsize=15, margin=0, padding=0):
        super(HeaderLabel, self).__init__(text, parent=parent)
        self.setMinimumHeight(height)
        self.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.setStyleSheet('background-color: {}; color: {};'
                           'font-weight: bold; font-size: {}px;'
                           'border-radius: {}px; margin: {}px; padding: {}px'
                           .format(bgcolor, color, fontsize,bsides,margin,padding))

class SubHeaderLabel(HeaderLabel):

    def __init__(self, text, parent=None, height=22, fontsize=14, color='#fefefe',
                 bgcolor='#009999'):
        super(SubHeaderLabel, self).__init__(text, parent=parent, height=height,
                                             bgcolor=bgcolor, color=color,
                                             bradius=3, bsides=2,
                                             fontsize=fontsize)


class HeaderButton(object):

    def __init__(self, parent, index, enabled=True):
        self.parent = parent
        self.index = index
        self.setEnabled(enabled)

    def setEnabled(self, bool):
        self.parent.setTabEnabled(self.index, bool)

    def setDisabled(self, bool):
        self.parent.setTabDisabled(self.index, bool)

    def setVisible(self, bool):
        #self.parent.setCurrentIndex(self.index)
        pass

class MHBoxLayout(QtWidgets.QHBoxLayout):
    def __init__(self, *widgets, **kwargs):
        super(MHBoxLayout, self).__init__()
        self.setContentsMargins(0,0,0,0)

class MVBoxLayout(QtWidgets.QVBoxLayout):
    def __init__(self, *widgets, **kwargs):
        super(MVBoxLayout, self).__init__()
        self.setContentsMargins(0,0,0,0)

class HWidgets(QtWidgets.QWidget):

    def __init__(self, *widgets, **kwargs):
        kwargs.setdefault('parent', None)
        super(HWidgets, self).__init__(kwargs['parent'])

        s = [0]*len(widgets) if not 'stretch' in kwargs else kwargs['stretch']
        if len(s) < len(widgets):
            s += [0] * (len(widgets) - len(s))

        hbox = QtWidgets.QHBoxLayout(self)
        self.setLayout(hbox)
        hbox.setContentsMargins(0,0,0,0)
        for i, widget in enumerate(widgets):
            if widget is None:
                widget = QtWidgets.QWidget()
            elif isinstance(widget, six.string_types):
#            elif type(widget) in [str, unicode]:
                widget = QtWidgets.QLabel(widget)
            hbox.addWidget(widget, s[i])
        self.widgets = widgets

    def setStyle(self, style):
        for widget in self.widgets:
            widget.setStyleSheet(style)
        self.setStyleSheet(style)

class BLabel(QtWidgets.QLabel):

    def __init__(self, *args, **kwargs):
        super(BLabel, self).__init__(*args, **kwargs)
        self.setStyleSheet('font-weight: bold;')

class RLabel(QtWidgets.QLabel):

    def __init__(self, *args, **kwargs):
        super(RLabel, self).__init__(*args, **kwargs)
        self.setAlignment(QtCore.Qt.AlignRight)

class ULabel(QtWidgets.QLabel):

    def __init__(self, *args, **kwargs):
        super(ULabel, self).__init__(*args, **kwargs)
        self.setStyleSheet('color: #6194BC;')

class TComboBox(QtWidgets.QWidget):

    def __init__(self, label, options, selected=None, parse=str, default=0, *args, **kwargs):
        super(TComboBox, self).__init__(*args, **kwargs)
        self.default = default
        hbox = MHBoxLayout()
        if label is not None:
            hbox.addWidget(QtWidgets.QLabel(label))
        self.combo = SComboBox()
        for option in options:
            self.combo.addItem(str(option))
        if selected:
            self.combo.setCurrentIndex(selected)
        hbox.addWidget(self.combo, 1)
        self.setLayout(hbox)
        self.currentIndexChanged = self.combo.currentIndexChanged
        self.parse = parse
        self.findText = self.combo.findText

    def clear(self):
        self.combo.clear()

    def currentIndex(self):
        return self.combo.currentIndex()

    def currentText(self):
        return str(self.combo.currentText())

    def setCurrentIndex(self, idx):
        self.combo.setCurrentIndex(idx)

    def addItem(self, item):
        self.combo.addItem(item)

    def removeItem(self, item):
        self.combo.removeItem(item)

    def updateItems(self, items):
        self.combo.clear()
        for option in items:
            self.combo.addItem(option)

    def count(self):
        return self.combo.count()

    def itemText(self, i):
        return self.combo.itemText(i)

    def setItemText(self, i, txt):
        self.combo.setItemText(i, txt)

    def value(self):
        try:
            return self.parse(self.combo.currentText())
        except:
            return self.parse(self.default)

class PLineEdit(QtWidgets.QLineEdit):

    def __init__(self, default=0, parse=int, fontsize=None, *args, **kwargs):
        super(PLineEdit, self).__init__(*args, **kwargs)
        self.default = default
        self.parse = parse
        self.setPlaceholderText(str(default))

        if fontsize is not None:
            self.setStyleSheet('font-size: {}px;'.format(fontsize))

    def value(self):
        val = self.default
        try:
            val = self.parse(self.text())
        except:
            pass
        return val

    def setDefault(self, val):
        self.default = val
        self.setPlaceholderText(str(val))


class RCheckBox(QtWidgets.QWidget):

    def __init__(self, text, fontsize=None, **kwargs):
        super(RCheckBox, self).__init__(**kwargs)

        hbox = MHBoxLayout()
        self.setLayout(hbox)

        self.label = RClickableLabel(text)
        self.check = QtWidgets.QCheckBox()
        self.label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.check.setChecked(True)
        hbox.addWidget(self.label, 1)
        hbox.addWidget(self.check)
        self.label.clicked.connect(self.on_label)
        self.stateChanged = self.check.stateChanged

        if fontsize is not None:
            self.label.setStyleSheet('font-size: {}px;'.format(fontsize))

    def on_label(self):
        self.check.toggle()

    def setChecked(self, bol):
        self.check.setChecked(bol)

    def setCheckState(self, st):
        self.check.setCheckState(st)

    def isChecked(self):
        return self.check.isChecked()

    def setMaximumHeight(self, h):
        self.label.setMaximumHeight(h)
        self.check.setMaximumHeight(h)

    def setCheckStyle(self, style):
        self.check.setStyleSheet(style)

class RClickableLabel(RLabel):
    clicked = QtCore.pyqtSignal()
    def mousePressEvent(self, ev):
        self.clicked.emit()

class ClickableLabel(QtWidgets.QLabel):
    clicked = QtCore.pyqtSignal()
    def mousePressEvent(self, ev):
        self.clicked.emit()

class LCheckBox(QtWidgets.QWidget):

    def __init__(self, text, fontsize=None, **kwargs):
        super(LCheckBox, self).__init__(**kwargs)

        hbox = MHBoxLayout()
        self.setLayout(hbox)

        self.label = ClickableLabel(text)
        self.check = QtWidgets.QCheckBox()
        self.check.setMaximumWidth(20)
        self.check.setChecked(True)
        hbox.addWidget(self.check)
        hbox.addWidget(self.label, 1)
        self.label.clicked.connect(self.on_label)
        self.stateChanged = self.check.stateChanged

        if fontsize is not None:
            self.label.setStyleSheet('font-size: {}px;'.format(fontsize))

    def on_label(self):
        self.check.toggle()

    def setChecked(self, bol):
        self.check.setChecked(bol)

    def setCheckState(self, st):
        self.check.setCheckState(st)

    def isChecked(self):
        return self.check.isChecked()

    def setMaximumHeight(self, h):
        self.label.setMaximumHeight(h)
        self.check.setMaximumHeight(h)

    def setCheckStyle(self, style):
        self.check.setStyleSheet(style)

class HLayer(QtWidgets.QWidget):

    valueChanged = QtCore.pyqtSignal(str, str, int)
    toggled  = QtCore.pyqtSignal(str, str, bool)
    export = QtCore.pyqtSignal(str, str)

    def __init__(self, name, level, current=None, visible=True, **kwargs):
        super(HLayer, self).__init__(**kwargs)
        hbox = MHBoxLayout()
        self.setLayout(hbox)

        self.name = name
        self.level = level

        self.label = QtWidgets.QLabel(name + ':')
        self.label.setFixedWidth(90)
        self.label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.slider = CSlider(self.name, current=current)
        self.slider.lbl_current.setFixedWidth(25)
        self.check = QtWidgets.QCheckBox()
        self.check.setChecked(visible)
        #self.btn_tiff = QtWidgets.QPushButton('.rec')
        #self.btn_tiff.setFixedWidth(40)

        hbox.addWidget(self.label)
        hbox.addWidget(self.slider, 1)
        hbox.addWidget(self.check)
        #hbox.addWidget(self.btn_tiff)

        self.slider.valueChanged.connect(self.on_value)
        self.check.stateChanged.connect(self.on_toggled)
        #self.btn_tiff.clicked.connect(self.on_tiff_clicked)

    def on_value(self, val):
        self.valueChanged.emit(self.name, self.level, val)

    def on_toggled(self, val):
        self.toggled.emit(self.name, self.level, val)

    def on_tiff_clicked(self):
        self.export.emit(self.name, self.level)

    def setChecked(self, value):
        self.check.setChecked(value)


class FLineEdit(QtWidgets.QLineEdit):

    strChanged = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        super(FLineEdit, self).__init__(*args, **kwargs)
        self.returnPressed.connect(self.clearFocus)

    def focusOutEvent(self, ev):
        super(FLineEdit, self).focusOutEvent(ev)
        self.strChanged.emit(self.text())

class ColorButton(QtWidgets.QPushButton):

    colorChanged = QtCore.pyqtSignal(str)

    def __init__(self, color='#000000', clickable=True, *args, **kwargs):
        super(ColorButton, self).__init__(*args, **kwargs)
        self.setColor(color)
        if clickable:
            self.clicked.connect(self.on_click)

    def c2h(self, color):
        return str(color.name())

    def setColor(self, color):
        if color is None:
            self.setStyleSheet('QPushButton, QPushButton:hover'
                               '{background-color:'
                               'qlineargradient(x1:0, y1:0, x2:1, y2:1,'
                               'stop: 0 white, stop: 0.15 white,'
                               'stop: 0.2 red,'
                               'stop: 0.25 white, stop: 0.45 white,'
                               'stop: 0.5 red,'
                               'stop: 0.55 white, stop: 0.75 white,'
                               'stop: 0.8 red, stop: 0.85 white);}')
        else:
            self.setStyleSheet('QPushButton {background-color: %s;}'
                               'QPushButton:hover {background-color:'
                               'qlineargradient(x1:0, y1:0, x2:0.5, y2:1,'
                               'stop: 0 white, stop: 1 %s);}'
                               % (color, color))
        self.color = color

    def on_click(self):
        c = QtWidgets.QColorDialog.getColor(QtGui.QColor(self.color), self.parent())
        if not c.isValid():
            return
        self.setColor(self.c2h(c))
        self.colorChanged.emit(self.color)

class HLabelParent(QtWidgets.QWidget):

    def __init__(self, name, **kwargs):
        super(HLabelParent, self).__init__(**kwargs)
        hbox = MHBoxLayout()
        self.setLayout(hbox)

        self.txt_label_name = QtWidgets.QLabel(name)
        self.combo_parent = SComboBox()
        self.combo_parent.addItem('None')
        dummy = HWidgets(self.txt_label_name, self.combo_parent,
                         stretch=[0,0,0])
        hbox.addWidget(dummy)

    def setColor(self, color):
        self.txt_label_name.setStyleSheet('background-color: %s;' % color)

    def setText(self, name):
        self.txt_label_name.setText(name)

    def addParent(self, name):
        self.combo_parent.addItem(name)

    def getParent(self):
        return str(self.combo_parent.currentText())


class HEditLabel(QtWidgets.QWidget):

    selected = QtCore.pyqtSignal(int)
    removed = QtCore.pyqtSignal(int)
    nameChanged = QtCore.pyqtSignal(int, str)
    colorChanged = QtCore.pyqtSignal(int, str)
    parentRequested = QtCore.pyqtSignal(int)
    visibilityChanged = QtCore.pyqtSignal(int, bool)

    def __init__(self, name, idx, visible=True, **kwargs):
        super(HEditLabel, self).__init__(**kwargs)
        hbox = MHBoxLayout()
        self.setLayout(hbox)

        self.idx = idx

        self.btn_remove = QtWidgets.QPushButton('X')
        self.btn_remove.setContentsMargins(0,0,0,0)
        self.btn_remove.setMaximumWidth(25)
        self.btn_remove.setCheckable(True)

        self.label = FLineEdit(name)
        self.label.setMinimumWidth(100)
        self.label.setMaximumWidth(100)

        self.lbutton = ColorButton()
        self.pbutton = ColorButton(color=None)
        self.pbutton.clicked.disconnect()

        self.sbutton = QtWidgets.QPushButton(u'\u2713')
        self.sbutton.setContentsMargins(0,0,0,0)
        self.sbutton.setMaximumWidth(25)
        self.sbutton.setCheckable(True)

        self.check = QtWidgets.QCheckBox()
        self.check.setChecked(True)

        hbox.addWidget(self.btn_remove)
        hbox.addWidget(self.label)
        hbox.addWidget(self.lbutton, 1)
        hbox.addWidget(self.pbutton, 1)
        hbox.addWidget(self.sbutton)
        #hbox.addWidget(self.check)

        self.sbutton.clicked.connect(self.on_select)
        self.label.strChanged.connect(self.on_name)
        self.lbutton.colorChanged.connect(self.on_color)
        self.pbutton.clicked.connect(self.on_parent_clicked)
        self.check.stateChanged.connect(self.on_check)
        self.btn_remove.clicked.connect(self.on_remove)

    def on_name(self, name):
        self.nameChanged.emit(self.idx, name)

    def on_color(self, color):
        self.colorChanged.emit(self.idx, color)

    def on_select(self):
        self.selected.emit(self.idx)

    def on_check(self, val):
        self.visibilityChanged.emit(self.idx, val)

    def on_remove(self, val):
        self.removed.emit(self.idx)

    def setName(self, value):
        self.label.setText(value)

    def setColor(self, value):
        self.lbutton.setColor(value)

    def setChecked(self, bol):
        self.check.setChecked(bol)

    def setSelected(self, bol):
        self.btn_remove.setChecked(bol)
        self.sbutton.setChecked(bol)

    def on_parent_clicked(self):
        self.parentRequested.emit(self.idx)

    def setParentColor(self, color):
        self.pbutton.setColor(color)


class CSlider(QtWidgets.QWidget):

    valueChanged = QtCore.pyqtSignal(int)

    def __init__(self, name, vmin=0, vmax=100, current=None,
                 interval=1, lblwidth=50, **kwargs):
        super(CSlider, self).__init__(**kwargs)

        vbox = MVBoxLayout()
        self.setLayout(vbox)

        self.current = current if current is not None else vmax
        self.vmin = vmin
        self.vmax = vmax
        self.interval = interval
        self.initial = self.current

        # Header
        self.lbl_current = QtWidgets.QLabel("%d" % self.current)
        self.lbl_current.setMinimumWidth(lblwidth)
        self.lbl_current.setStyleSheet('color: #6194BC;')

        # Slider
        self.sld_val = QtWidgets.QSlider(1)
        self.sld_val.setMinimum(vmin)
        self.sld_val.setMaximum(vmax)
        self.sld_val.setValue(self.current)
        self.sld_val.setTickInterval(interval)
        self.sld_val.setSingleStep(interval)
        self.sld_val.setTracking(True)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.sld_val, 1)
        hbox.addWidget(self.lbl_current)
        vbox.addLayout(hbox)

        # Slots
        self.sld_val.valueChanged.connect(self.on_value_changed)
        self.sld_val.wheelEvent = self.wheelEvent

    def on_value_changed(self, val):
        self.current = val
        self.lbl_current.setText("%d" % self.current)
        self.valueChanged.emit(self.current)

    def value(self):
        return self.current

    def setValue(self, val):
        self.sld_val.setValue(val)

    def maximum(self):
        return self.sld_val.maximum()

    def minimum(self):
        return self.sld_val.minimum()

    def wheelEvent(self, e):
        if hasattr(e, 'delta'):
            wheel_delta = e.delta()
        else:
            wheel_delta = e.angleDelta().y()
        if wheel_delta > 0 and self.value() < self.maximum():
            self.setValue(self.value()+self.interval)
        elif wheel_delta < 0 and self.value() > self.minimum():
            self.setValue(self.value()-self.interval)


class OddSlider(CSlider):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('interval', 2)
        super(OddSlider, self).__init__(*args, **kwargs)

    def on_value_changed(self, val):
        if val%2 == 0:
            if val < self.maximum():
                val = val + 1
            else:
                val = val - 1
        self.current = val
        self.lbl_current.setText("%d" % self.current)
        self.valueChanged.emit(self.current)


class WSlider(QtWidgets.QSlider):

    def __init__(self, maximum, minimum=0, orient=1, parent=None):
        super(WSlider, self).__init__(orient, parent=parent)

        self.setMinimum(minimum)
        self.setMaximum(maximum)
        self.setValue(maximum//2)
        self.setSingleStep(1)

    def wheelEvent(self, e):
        if hasattr(e, 'delta'):
            wheel_delta = e.delta()
        else:
            wheel_delta = e.angleDelta().y()
        if wheel_delta > 0 and self.value() < self.maximum():
            self.setValue(self.value()+1)
        elif wheel_delta < 0 and self.value() > self.minimum():
            self.setValue(self.value()-1)

class WSliderPane(QtWidgets.QWidget):

    def __init__(self,  maximum, minimum=0, orient=1, parent=None):
        super(WSliderPane, self).__init__(parent=parent)

        hbox = QtWidgets.QHBoxLayout()
        self.setLayout(hbox)
        self.slider = WSlider(maximum, minimum=minimum, orient=orient)

        hbox.addWidget(self.slider, 1)

        self.valueChanged = self.slider.valueChanged
        self.setMinimum = self.slider.setMinimum
        self.setMaximum = self.slider.setMaximum
        self.setValue = self.slider.setValue

        self.wheelEvent = self.slider.wheelEvent
        self.value = self.slider.value
        self.setValue = self.slider.setValue
        self.maximum = self.slider.maximum
        self.minimum = self.slider.minimum
        self.setTracking = self.slider.setTracking

class HSlider(QtWidgets.QWidget):

    valueChanged = QtCore.pyqtSignal(float)

    def __init__(self, name, vmin, vmax, current=None, **kwargs):
        super(HSlider, self).__init__(**kwargs)

        vbox = MVBoxLayout()
        self.setLayout(vbox)

        self.current = current
        if self.current is None:
            self.current = (vmax - vmin) / 2.
        self.vmin = vmin
        self.vmax = vmax

        # Header
        hbox = MHBoxLayout()
        self.lbl_name = QtWidgets.QLabel(name)
        self.lbl_name.setStyleSheet('font-weight: bold;')
        self.lbl_current = QtWidgets.QLabel("%.2f" % self.current)
        self.lbl_current.setStyleSheet('color: #6194BC;')

        hbox.addWidget(self.lbl_name)
        hbox.addWidget(self.lbl_current)
        vbox.addLayout(hbox)

        # Slider
        self.lbl_min = QtWidgets.QLabel('%.2f' % vmin)
        self.lbl_min.setMinimumWidth(40)
        self.lbl_min.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_max = QtWidgets.QLabel('%.2f' % vmax)
        self.lbl_max.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_max.setMinimumWidth(40)
        self.sld_val = QtWidgets.QSlider(1)
        self.sld_val.setMinimum(0)
        self.sld_val.setMaximum(1000)
        self.sld_val.setValue(self.val2slider(self.current))

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.lbl_min)
        hbox.addWidget(self.sld_val)
        hbox.addWidget(self.lbl_max)
        vbox.addLayout(hbox)

        # Slots
        self.sld_val.valueChanged.connect(self.on_value_changed)

    def val2slider(self, val):
        return (val - self.vmin) / float(self.vmax - float(self.vmin)) * 1000

    def slider2val(self, val):
        return val * (self.vmax - float(self.vmin)) / 1000. + self.vmin

    def on_value_changed(self, val):
        self.current = self.slider2val(val)
        self.lbl_current.setText("%.2f" % self.current)
        self.valueChanged.emit(self.current)

    def value(self):
        return self.current

    def setMinimum(self, val):
        self.vmin = val
        self.lbl_min.setText('%.2f' % val)

    def setMaximum(self, val):
        self.vmax = val
        self.lbl_max.setText('%.2f' % val)

    def setValue(self, val):
        self.sld_val.setValue(int(self.val2slider(float(val))))

class HSize3D(QtWidgets.QWidget):

    def __init__(self, name, default=(0, 0, 0), txtwidth=None,
                 parse=int, coordwidth=None):
        super(HSize3D, self).__init__()

        self.default = default
        self.parse = parse

        hbox = MHBoxLayout()
        self.setLayout(hbox)

        self.zbox = QtWidgets.QLineEdit()
        self.ybox = QtWidgets.QLineEdit()
        self.xbox = QtWidgets.QLineEdit()

        if coordwidth is not None:
            self.zbox.setMaximumWidth(coordwidth)
            self.ybox.setMaximumWidth(coordwidth)
            self.xbox.setMaximumWidth(coordwidth)

        self.zbox.setPlaceholderText(str(default[0]))
        self.ybox.setPlaceholderText(str(default[1]))
        self.xbox.setPlaceholderText(str(default[2]))

        zlbl = QtWidgets.QLabel('z')
        zlbl.setStyleSheet('color: #6194BC;')
        ylbl = QtWidgets.QLabel('y')
        ylbl.setStyleSheet('color: #6194BC;')
        xlbl = QtWidgets.QLabel('x')
        xlbl.setStyleSheet('color: #6194BC;')

        title = QtWidgets.QLabel(name + ':')

        if txtwidth is not None:
            title.setMinimumWidth(txtwidth)
            title.setMaximumWidth(txtwidth)

        hbox.addWidget(title)
        hbox.addWidget(zlbl)
        hbox.addWidget(self.zbox, 1)
        hbox.addWidget(ylbl)
        hbox.addWidget(self.ybox, 1)
        hbox.addWidget(xlbl)
        hbox.addWidget(self.xbox, 1)

    def value(self):
        z, y, x = self.default
        try:
            z = self.parse(self.zbox.text())
        except:
            pass
        try:
            y = self.parse(self.ybox.text())
        except:
            pass
        try:
            x = self.parse(self.xbox.text())
        except:
            pass
        return (z, y, x)

    def setValue(self, value):
        self.zbox.setText(str(value[0]))
        self.ybox.setText(str(value[1]))
        self.xbox.setText(str(value[2]))




import numpy as np
import logging as log

import networkx as nx
from sklearn.decomposition import PCA, IncrementalPCA
from sklearn.preprocessing import StandardScaler
from sklearn.random_projection import SparseRandomProjection
from sklearn.ensemble import ExtraTreesClassifier, \
                             RandomForestClassifier, \
                             AdaBoostClassifier,\
                             GradientBoostingClassifier
from sklearn.svm import SVC
from sklearn.cluster import MiniBatchKMeans

from sklearn.linear_model import SGDClassifier

from ..lib._spencoding import _sp_labels
from ..lib.spencoding import spmeans, sphist, spstats
from ..lib.spgraph import aggregate_neighbors
from ..lib._qpbo import solve_binary, solve_aexpansion

from ..core import DataModel


DM = DataModel.instance()


def obtain_classifier(clf_p):
    if clf_p['clf'] == 'ensemble':
        mode = 'ensemble'
        if clf_p['type'] == 'rf':
            clf = RandomForestClassifier(n_estimators=clf_p['n_estimators'],
                                         max_depth=clf_p['max_depth'],
                                         n_jobs=clf_p['n_jobs'])
        elif clf_p['type'] == 'erf':
            clf = ExtraTreesClassifier(n_estimators=clf_p['n_estimators'],
                                       max_depth=clf_p['max_depth'],
                                       n_jobs=clf_p['n_jobs'])
        elif clf_p['type'] == 'ada':
            clf = AdaBoostClassifier(n_estimators=clf_p['n_estimators'],
                                     learning_rate=clf_p['learning_rate'])
        else:
            clf = GradientBoostingClassifier(n_estimators=clf_p['n_estimators'],
                                             max_depth=clf_p['max_depth'],
                                             learning_rate=clf_p['learning_rate'],
                                             subsample=clf_p['subsample'])
    elif clf_p['clf'] == 'svm':
        mode = 'svm'
        clf = SVC(C=clf_p['C'], gamma=clf_p['gamma'], kernel=clf_p['kernel'],
                  probability=True)
    elif clf_p['clf'] == 'sgd':
        mode = 'sgd'
        clf = SGDClassifier(loss=clf_p['loss'], penalty=clf_p['penalty'],
                            alpha=clf_p['alpha'], n_iter=clf_p['n_iter'])
    else:
        raise Exception('Classifier not supported')
    return clf, mode


def compute_supervoxel_descriptor(supervoxels, descriptors, desc_type, desc_bins):
    'Mean' 'Quantized' 'Textons' 'Covar' 'Sigma Set'

    if desc_type == 'Mean':
        return spmeans(descriptors, supervoxels)
    elif desc_type == 'Covar':
        return spstats(descriptors, supervoxels, mode='add', norm=None)
    elif desc_type == 'Sigma Set':
        return spstats(descriptors, supervoxels, mode='add', sigmaset=True, norm=None)

    if desc_type == 'Textons':
        log.info('+ Applying PCA')
        descriptors = IncrementalPCA(batch_size=100).fit_transform(descriptors)

    log.info('+ Quantizing descriptors')
    cluster = MiniBatchKMeans(n_clusters=desc_bins, batch_size=100).fit_predict(descriptors)
    return sphist(cluster.astype(np.int32), supervoxels, nbins=desc_bins)


def extract_descriptors(supervoxels=None, features=None,
                        projection=None, desc_type=None, desc_bins=None,
                        nh_order=None, sp_edges=None):
    total = len(features)

    log.info('+ Reserving memory for {} features'.format(total))
    descriptors = np.zeros(DM.region_shape() + (total,), np.float32)

    for i in range(total):
        log.info('    * Loading feature {}'.format(features[i]))
        descriptors[..., i] = DM.load_slices(features[i])

    sp = None
    mask = None

    if supervoxels is not None:
        log.info('+ Loading supervoxels')
        sp = DM.load_slices(supervoxels)
        if sp.min() < 0:
            raise Exception('Supervoxels need to be recomputed for this ROI')
        descriptors.shape = (-1, total)

        num_sv = DM.attr(supervoxels, 'num_supervoxels')
        mask = np.zeros(num_sv, np.bool)
        mask[sp.ravel()] = True

        log.info('+ Computing descriptors: {} ({})'.format(desc_type, desc_bins))
        descriptors = compute_supervoxel_descriptor(sp, descriptors,
                                                    desc_type, desc_bins)
        nh_order = int(nh_order)
        if nh_order > 0:
            log.info('+ Loading edges into memory')
            edges = DM.load_ds(sp_edges)

            log.info('+ Filtering edges for selected ROI')
            idx = mask[edges[:, 0]] & mask[edges[:, 1]]
            edges = edges[idx]

            log.info('+ Aggregating neighbour features')
            G = nx.Graph()
            G.add_edges_from(edges)
            descriptors = aggregate_neighbors(descriptors, G, mode='append',
                                              norm='mean', order=nh_order)
        descriptors = descriptors[mask]

    return descriptors, sp, mask


def predict_proba(y_data=None, p_data=None, train=False, append=False,
                  level_params=None, desc_params=None,
                  clf_params=None, ref_params=None,
                  out_labels=None, out_confidence=None):
    log.info("+ Extracting descriptors: {}".format(desc_params))
    X, supervoxels, svmask = extract_descriptors(**desc_params)
    full_svmask = svmask.copy()

    if level_params['plevel'] is not None:
        parent_labels = DM.load_slices(p_data)
        mask = parent_labels == level_params['plabel']
    else:
        mask = None

    if train:
        clf, mode = obtain_classifier(clf_params)
        log.info("+ Creating classifier: {}".format(clf_params))

        log.info("+ Loading labels")
        labels = DM.load_slices(y_data)
    else:
        if DM.has_classifier():
            log.info("+ Getting existing classifier")
            clf = DM.get_classifier_from_model()
        else:
            log.error("For some reason, no previous classifier exists!")
            return

    if mask is not None:
        mask = mask.astype(np.int16)
        mask = np.bincount(supervoxels.ravel(), weights=mask.ravel() * 2 - 1) > 0
        mask = mask[svmask]
        mask.shape = -1

    if supervoxels is not None:
        nsp = DM.attr(desc_params['supervoxels'], 'num_supervoxels')
        log.info('+ Extracting supervoxel labels')
        if train:
            nlbl = DM.attr(y_data, 'label').max() + 1
            labels = _sp_labels(supervoxels.ravel(), labels.ravel(), nsp, nlbl, 0)
            y = labels
            y.shape = -1

            if mask is not None:
                idx_train = (y > -1) & mask
            else:
                idx_train = (y > -1)

            X_train = X[idx_train]
            y_train = y[idx_train]
            log.info("+ Initial X_train shape {}".format(X_train.shape))
            log.info("+ Initial y_train shape {}".format(y_train.shape))

            if append and DM.has_training_data():
                previous_X, previous_y = DM.get_training_data()
                X_train = np.append(previous_X, X_train, axis=0)
                y_train = np.append(previous_y, y_train)
                log.info("+ Appended X_train shape {}".format(X_train.shape))
                log.info("+ Appended y_train shape {}".format(y_train.shape))

            DM.add_training_data(X_train, y_train)

            if clf is not None:
                _train_classifier(clf, X_train, y_train,
                                  project=desc_params['projection'])
                log.info('+ Adding classifier to data model')
                DM.add_classifier_to_model(clf, clf_params)
            elif clf is None:
                log.error("No Classifier found!")
                return None

    return predict_and_save(X, clf, full_svmask, nsp, out_confidence, out_labels, supervoxels, ref_params, mask, svmask)


def predict_and_save(X, clf, full_svmask, nsp, out_confidence, out_labels, supervoxels, ref_params, mask, svmask):
    result = classifier_predict(X, clf)
    probs = result['probs']
    labels = np.asarray(list(set(np.unique(result['class'][supervoxels])) - set([-1])), np.int32)
    pred = result['class']
    if supervoxels is not None:
        if ref_params['ref_type'] != 'None':
            log.info('+ Remapping supervoxels')
            svmap = np.full(nsp, -1, supervoxels.dtype)
            svmap[full_svmask] = np.arange(svmask.sum())

            log.info('+ Extracting graph')
            refine_lamda = ref_params['lambda']
            edges = DM.load_ds(ref_params['sp_edges'])
            edge_weights = DM.load_ds(ref_params['sp_eweights'])
            mean_weights = edge_weights.mean()
            edge_weights = np.minimum(edge_weights, mean_weights) / mean_weights

            log.info('+ Remapping edges')
            idx = full_svmask[edges[:, 0]] & full_svmask[edges[:, 1]]
            edges = edges[idx]
            edges[:, 0] = svmap[edges[:, 0]]
            edges[:, 1] = svmap[edges[:, 1]]
            edge_weights = edge_weights[idx]

            log.info('  * Unary potentials')
            unary = (-np.ma.log(probs)).filled()
            unary = unary.astype(np.float32)
            mapping = np.zeros(pred.max() + 1, np.int32)
            mapping[labels] = np.arange(labels.size)
            idx = np.where(pred > -1)[0]
            col = mapping[pred[idx]]
            unary[idx, col] = 0
            pairwise = edge_weights.astype(np.float32)

            log.info('  * Pairwise potentials')
            if ref_params['ref_type'] == 'Appearance':
                log.info('+ Extracting descriptive weights')
                dists = np.sqrt(np.sum((X[edges[:, 0]] - X[edges[:, 1]]) ** 2, axis=1))
                pairwise *= np.exp(-dists.mean() * dists)

            log.debug("  * Shapes: {}, {}, {}, {}".format(unary.shape, pairwise.shape,
                                                          edges.shape, edge_weights.shape))
            log.debug("  * Edges: min: {}, max: {}".format(edges.min(0), edges.max(0)))
            log.debug("  * Unary: min: {}, max: {}, mean: {}"
                      .format(unary.min(0), unary.max(0), unary.mean(0)))
            log.debug("  * Pairwise: min: {}, max: {}, mean: {}"
                      .format(pairwise.min(), pairwise.max(), pairwise.mean()))

            log.info('+ Refining labels')
            label_cost = np.ones((labels.size, labels.size), np.float32)
            label_cost[np.diag_indices_from(label_cost)] = 0
            label_cost *= refine_lamda
            if label_cost.shape[0] == 2:
                refined = solve_binary(edges, unary, pairwise, label_cost)
            else:
                refined = solve_aexpansion(edges, unary, pairwise, label_cost)
            pred[mask] = labels[refined[mask]]

            if mask is not None:
                pred[~mask] = -1
                #conf[~mask] = -1
        else:
            pass  # TODO pixel refinement
        log.info('+ Mapping predictions and probabilities back to pixels')
        pred_map = np.empty(nsp, dtype=result['class'].dtype)
        conf_map = np.empty(nsp, dtype=result['probs'].dtype)
        # Get a list of label classes since when labels have been deleted some will be missing
        label_classes = np.unique(result['class'])
        # Make a dict mapping from label to sequential numbers beginning from 0
        dict_map = dict(zip(label_classes, range(len(label_classes))))
        # Slice each list of confidences with the corresponding predicted class
        # to return the confidence for that class
        conf_result = list(map(lambda x, y: x[dict_map[y]], result['probs'], result['class']))
        pred_map[full_svmask] = result['class']
        conf_map[full_svmask] = conf_result
        if mask is not None:
            pred_map[~mask] = -1
            conf_map[~mask] = -1
        pred = pred_map[supervoxels]
        conf = conf_map[supervoxels]
        pred.shape = DM.region_shape()
        conf.shape = DM.region_shape()
    log.info('+ Saving results to disk')
    DM.create_empty_dataset(out_labels, shape=DM.data_shape, dtype=pred.dtype)

    DM.write_slices(out_labels, pred, params=dict(labels=labels, active=True))
    DM.create_empty_dataset(out_confidence, shape=DM.data_shape, dtype=conf.dtype)
    DM.write_slices(out_confidence, conf, params=dict(labels=labels, active=True))
    return out_labels, out_confidence, labels

def _train_classifier(clf, X_train, y_train, rnd=42, project=None):

    if project is not None and project != 'None':
        log.info('+ Projecting features')
        if project == 'random_projection':
            log.info('  * Sparse Random Projection')
            proj = SparseRandomProjection(n_components=X_train.shape[1], random_state=rnd)
        elif project == 'standard':
            log.info('  * Standard Projection')
            proj = StandardScaler()
        elif project == 'pca':
            log.info('  * Principle Component Analysis')
            proj = IncrementalPCA(batch_size=100)
        elif project == 'random_pca':
            log.info('  * Randomized Principle Component Analysis')
            proj = PCA(n_components=X_train.shape[1], svd_solver='randomized', whiten=True, random_state=rnd)
        else:
            log.error('Projection {} not available'.format(project))
            return

        X_train = proj.fit_transform(X_train)

    log.info('+ Training classifier')
    clf.fit(X_train, y_train)

def classifier_predict(X, clf):
    result = {}
    log.info('+ Predicting labels')
    result['class'] = clf.predict(X)
    result['probs'] = clf.predict_proba(X)
    return result





import numpy as np

from ._spencoding import _sphist, _spmeans, _spstats


def normalize(features, norm='l1'):
    if norm in ['l1', 'hellinger']:
        features /= np.abs(features).sum(axis=1)[:, None]
    elif norm == 'l2':
        features /= np.sqrt((features**2).sum(axis=1))[:, None]
    elif norm == 'linf':
        features /= np.abs(features).max(axis=1)[:, None]
    elif norm == 'unit':
        features -= features.min(0)
        features /= features.max(0)
    if norm == 'hellinger':
        np.sqrt(features, features) # inplace
    return features


def sphist(data, splabels, nbins=100, norm=None):
    features = _sphist(data, splabels.flatten(), splabels.max()+1, nbins)
    return normalize(features, norm=norm)


def spmeans(data, splabels, norm=None):
    features =  _spmeans(data, splabels.flatten(), splabels.max()+1)
    return normalize(features, norm=norm)


def spstats(data, splabels, mode='append', sigmaset=False, covmode='full', norm=None):

    if mode not in ['append', 'add', None]:
        raise Exception('Only `append` or `add` methods are accepted')

    means, covars = _spstats(data, splabels.flatten(), splabels.max()+1)

    if sigmaset:
        # Add small constant to covars to make them positive-definite
        covars += np.eye(covars.shape[-1])[None, ...] * 1e-5
        covars = np.linalg.cholesky(covars) * np.sqrt(means.shape[1])

        if covmode == 'full':
            y1, x1 = np.tril_indices(means.shape[1], k=-1)
            y2, x2 = np.triu_indices(means.shape[1], k=1)
            covars[:, y2, x2] = covars[:, y1, x1]

    if mode == 'add':
        covars += means[:, :, None]

    if sigmaset and covmode == 'tril':
        y, x = np.tril_indices(means.shape[1])
        covars = covars[:, y, x]
    else:
        covars.shape = (covars.shape[0], -1)

    if mode == 'append':
        features = np.c_[means, covars]
    else:
        features = covars

    return normalize(features, norm=norm)





import numpy as np
import networkx as nx

from ._rag import _create_rag_2d, _create_rag_3d
from skimage.segmentation import mark_boundaries

from skimage import measure, draw

from sklearn.metrics.pairwise import distance_metrics, kernel_metrics
from ._dist import bhattacharya, intersection


def edge_weight(x, y, mode='rbf', gamma=0.5):
    dists = distance_metrics()
    kernels = kernel_metrics()
    kernels['bhattacharya'] = bhattacharya
    kernels['intersection'] = intersection
    if mode in dists:
        diff = dists[mode](x, y)
    elif mode in kernels:
        diff = kernels[mode](x, y, gamma=gamma)
    else:
        raise Exception('Mode not recognised')

    return np.float64(diff)


def is_distance(mode):
    return mode in distance_metrics()


def min_weight(graph, src, dst, n):
    """Callback to handle merging nodes by choosing minimum weight.
    Returns either the weight between (`src`, `n`) or (`dst`, `n`)
    in `graph` or the minimum of the two when both exist.
    Parameters
    ----------
    graph : RAG
        The graph under consideration.
    src, dst : int
        The verices in `graph` to be merged.
    n : int
        A neighbor of `src` or `dst` or both.
    Returns
    -------
    weight : float
        The weight between (`src`, `n`) or (`dst`, `n`) in `graph` or the
        minimum of the two when both exist.
    """

    # cover the cases where n only has edge to either `src` or `dst`
    default = {'weight': np.inf}
    w1 = graph[n].get(src, default)['weight']
    w2 = graph[n].get(dst, default)['weight']
    return min(w1, w2)


class WRAG(nx.Graph):

    """
    The Region Adjacency Graph (RAG) of an image, subclasses
    `networx.Graph <http://networkx.github.io/documentation/latest/reference/classes.graph.html>`_
    """

    def __init__(self, data=None, **attr):
        super(WRAG, self).__init__(data, **attr)
        try:
            self.max_id = max(self.nodes_iter())
        except ValueError:
            # Empty sequence
            self.max_id = 0
        self.dist_weights = None

    def merge_nodes(self, src, dst, weight_func=min_weight, in_place=True,
                    extra_arguments=[], extra_keywords={}):
        """Merge node `src` and `dst`.
        The new combined node is adjacent to all the neighbors of `src`
        and `dst`. `weight_func` is called to decide the weight of edges
        incident on the new node.
        Parameters
        ----------
        src, dst : int
            Nodes to be merged.
        weight_func : callable, optional
            Function to decide edge weight of edges incident on the new node.
            For each neighbor `n` for `src and `dst`, `weight_func` will be
            called as follows: `weight_func(src, dst, n, *extra_arguments,
            **extra_keywords)`. `src`, `dst` and `n` are IDs of vertices in the
            RAG object which is in turn a subclass of
            `networkx.Graph`.
        in_place : bool, optional
            If set to `True`, the merged node has the id `dst`, else merged
            node has a new id which is returned.
        extra_arguments : sequence, optional
            The sequence of extra positional arguments passed to
            `weight_func`.
        extra_keywords : dictionary, optional
            The dict of keyword arguments passed to the `weight_func`.
        Returns
        -------
        id : int
            The id of the new node.
        Notes
        -----
        If `in_place` is `False` the resulting node has a new id, rather than
        `dst`.
        """
        src_nbrs = set(self.neighbors(src))
        dst_nbrs = set(self.neighbors(dst))
        neighbors = (src_nbrs | dst_nbrs) - set([src, dst])

        if in_place:
            new = dst
        else:
            new = self.next_id()
            self.add_node(new)

        for neighbor in neighbors:
            w = weight_func(self, src, new, neighbor, *extra_arguments,
                            **extra_keywords)
            self.add_edge(neighbor, new, weight=w)

        self.node[new]['labels'] = (self.node[src]['labels'] +
                                    self.node[dst]['labels'])
        self.remove_node(src)

        if not in_place:
            self.remove_node(dst)

        return new

    def add_node(self, n, attr_dict=None, **attr):
        """Add node `n` while updating the maximum node id.
        .. seealso:: :func:`networkx.Graph.add_node`."""
        super(RAG, self).add_node(n, attr_dict, **attr)
        self.max_id = max(n, self.max_id)

    def add_edge(self, u, v, attr_dict=None, **attr):
        """Add an edge between `u` and `v` while updating max node id.
        .. seealso:: :func:`networkx.Graph.add_edge`."""
        super(RAG, self).add_edge(u, v, attr_dict, **attr)
        self.max_id = max(u, v, self.max_id)

    def copy(self):
        """Copy the graph with its max node id.
        .. seealso:: :func:`networkx.Graph.copy`."""
        g = super(RAG, self).copy()
        g.max_id = self.max_id
        return g

    def next_id(self):
        """Returns the `id` for the new node to be inserted.
        The current implementation returns one more than the maximum `id`.
        Returns
        -------
        id : int
            The `id` of the new node to be inserted.
        """
        return self.max_id + 1

    def _add_node_silent(self, n):
        """Add node `n` without updating the maximum node id.
        This is a convenience method used internally.
        .. seealso:: :func:`networkx.Graph.add_node`."""
        super(RAG, self).add_node(n)

    def set_edge_weights(self, X, mode='rbf', gamma=0.5):
        for p, q, d in self.edges_iter(data=True):
            d['weight'] = edge_weight(X[p], X[q], mode=mode, gamma=gamma)
        self.dist_weights = is_distance(mode)


def rag_from_neighbors(nodes, neighbors, min_boundary=None,
                       norm_counts='unit', margin=0, return_rag=True):
    """Creates a Region Adjacency Graph between superpixels.

    Parameters
    ----------
    splabels : 2D or 3D ndarray
        Input superpixel labels for a 2D or 3D image.
    connectivity : int, optional
        Order of the neighbours, which must have a value of ${1, 2}$.
        For a 2D image `1` corresponds to 4-connected pixels while `2`
        corresponds to 8-connected pixels. For a 3D image `1` and `2`
        correspond to 6-connected and 18-connected voxels respectively.
    min_boundary : int, optional
        Minimum number of boundary pixels that a pair of superpixels
        need to share in order to consider them neighbours.
    norm_counts : string, optional
        Normalize the number of boundary pixels that any pair of
        superpixels share. Possible values are 'unit' which performs
        unit normalization as `counts /= counts.max()` and
        `margin` which performs `counts = np.max(counts, margin) / margin`.
    margin : int, optional
        An heuristic number of boundary pixels required to consider two
        neighbouring superpixels *perfect neighbours*. Used only if
        `norm_counts='margin'`.

    Returns
    -------
    graph : RAG (subclass of networkx.Graph)
        An undirected graph containing all the superpixels as nodes
        and the connections between them as edges. The number of
        boundary pixels that a pair of superpixels share is returned
        as an edge property: `boundary`.

    """
    n_nodes = np.int64(nodes.max() + 1)

    nodes = np.tile(nodes, neighbors.shape[1])
    neighbors = neighbors.flatten('f')
    neighbors[neighbors > -1] = nodes[neighbors[neighbors > -1]]

    idx = (neighbors != -1) & (neighbors != nodes)
    nodes = nodes[idx]
    neighbors = neighbors[idx]

    idx = nodes > neighbors
    nodes[idx], neighbors[idx] = neighbors[idx], nodes[idx]

    crossing_hash = nodes + neighbors.astype(np.int64) * n_nodes
    if min_boundary is not None:
        unique_hash, counts = np.unique(crossing_hash, return_counts=True)
    else:
        unique_hash = np.unique(crossing_hash)

    neighbors = np.c_[unique_hash % n_nodes, unique_hash // n_nodes]
    neighbors = neighbors.astype(np.int32)

    if min_boundary is not None:
        idx = (counts >= min_boundary)
        neighbors = neighbors[idx]
        counts = counts[idx]
        if norm_counts == 'unit':
            counts /= float(counts.max())
        elif norm_counts == 'margin':
            counts = np.minimum(counts, margin) / float(margin)
    else:
        counts = np.ones(neighbors.shape[0])

    # Create Region Adjacency Graph
    if return_rag:
        graph = WRAG()
        graph.add_nodes_from(np.arange(n_nodes))
        graph.add_weighted_edges_from(np.c_[neighbors, counts], weight='boundary')
        return graph

    return neighbors


def create_rag(splabels, connectivity=1, min_boundary=None,
               norm_counts='unit', margin=0, return_rag=True,
               return_counts=True):
    """Creates a Region Adjacency Graph between superpixels.

    Parameters
    ----------
    splabels : 2D or 3D ndarray
        Input superpixel labels for a 2D or 3D image.
    connectivity : int, optional
        Order of the neighbours, which must have a value of ${1, 2}$.
        For a 2D image `1` corresponds to 4-connected pixels while `2`
        corresponds to 8-connected pixels. For a 3D image `1` and `2`
        correspond to 6-connected and 18-connected voxels respectively.
    min_boundary : int, optional
        Minimum number of boundary pixels that a pair of superpixels
        need to share in order to consider them neighbours.
    norm_counts : string, optional
        Normalize the number of boundary pixels that any pair of
        superpixels share. Possible values are 'unit' which performs
        unit normalization as `counts /= counts.max()` and
        `margin` which performs `counts = np.max(counts, margin) / margin`.
    margin : int, optional
        An heuristic number of boundary pixels required to consider two
        neighbouring superpixels *perfect neighbours*. Used only if
        `norm_counts='margin'`.

    Returns
    -------
    graph : RAG (subclass of networkx.Graph)
        An undirected graph containing all the superpixels as nodes
        and the connections between them as edges. The number of
        boundary pixels that a pair of superpixels share is returned
        as an edge property: `boundary`.

    """
    n_labels = splabels.max() + 1

    if (splabels.ndim == 2 and connectivity not in (4, 8)) or \
       (splabels.ndim == 3 and connectivity not in (6, 18, 26)):
        raise Exception('Only {1, 2} values are supported for `connectivity`')

    if splabels.ndim == 2:
        nodes, neighbors = _create_rag_2d(splabels, connectivity)
    else:
        nodes, neighbors = _create_rag_3d(splabels, connectivity)

    nodes = np.tile(nodes, connectivity//2)
    neighbors = neighbors.flatten('f')

    idx = (neighbors != -1) & (neighbors != nodes)
    nodes = nodes[idx]
    neighbors = neighbors[idx]

    idx = nodes > neighbors
    nodes[idx], neighbors[idx] = neighbors[idx], nodes[idx]

    n_nodes = np.int64(splabels.max()+1)
    crossing_hash = nodes + neighbors.astype(np.int64) * n_nodes
    if min_boundary is not None or return_counts:
        unique_hash, counts = np.unique(crossing_hash, return_counts=True)
    else:
        unique_hash = np.unique(crossing_hash)

    neighbors = np.c_[unique_hash % n_nodes, unique_hash // n_nodes]
    neighbors = neighbors.astype(np.int32)

    if min_boundary is not None:
        idx = (counts >= min_boundary)
        neighbors = neighbors[idx]
        counts = counts[idx]

    if return_counts:
        if norm_counts == 'unit':
            counts /= float(counts.max())
        elif norm_counts == 'margin':
            counts = np.minimum(counts, margin) / float(margin)

    # Create Region Adjacency Graph
    if return_rag:
        graph = WRAG()
        graph.add_nodes_from(np.arange(n_labels))
        if return_counts:
            graph.add_weighted_edges_from(np.c_[neighbors, counts], weight='boundary')
        else:
            graph.add_edges_from(neighbors)
        return graph
    elif return_counts:
        return neighbors, counts
    else:
        return neighbors



def draw_rag(rag, splabels, img, border_color=(1,1,0),
             node_color=(0,0,1), edge_color=(0,1,0)):
    """Draws Region Adjacency Graph's nodes and edges in an image.

    Parameters
    ----------
    rag : RAG (subclass of networkx.Graph)
        Input graph of superpixels (i.e. created by `create_rag`).
    splabels : 2D ndarray
        Input superpixel labels for a 2D grayscale or color image.
    img : 2D or 3D ndarray
        2D grayscale or color input image.
    border_color : tuple of 3 int
        Color of the superpixel boundaries.
    node_color : tuple of 3 int
        Color of the nodes' centroids.
    edge_color : tuple of 3 int
        Color of edges between connected nodes.

    Returns
    -------
    out : color image, same shape as input `img`
        Returns the input image `img` with superpixel boundaries
        and graph's nodes and edges overlayed on top of it.

    """

    assert splabels.ndim == 2, 'Only 2D (+color) images are accepted'

    if img.ndim == 2:
        img = color.gray2rgb(img)

    regions = measure.regionprops(splabels+1)

    for n, region in enumerate(regions):
        rag.node[n]['centroid'] = region['centroid']

    if border_color is not None:
        img = mark_boundaries(img, splabels, color=border_color,
                              outline_color=None)

    for n1, n2, data in rag.edges_iter(data=True):
        r1, c1 = map(int, rag.node[n1]['centroid'])
        r2, c2 = map(int, rag.node[n2]['centroid'])

        circle = draw.circle(r1, c1, 2)
        img[circle] = node_color

        rr, cc = draw.line(r1, c1, r2, c2)
        img[rr, cc] = edge_color

    return img
